#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include "typedefs.h"

#define FIVE_MB (5 * 1024 * 1024)

void myError(char *msg)
{
    fprintf(stderr, "Error: %s\n", msg);
    exit(1);
}

/* convert a node list back to SML/NG */
void NodeToSMLNG(NODE *node, FILE *fp)
{
	char *outbuf = NULL;
	switch(node->type)
	{
	case TEXT:
		outbuf = node->s;
		break;
	case BOLD:
		outbuf = "<B>";
		break;
	case BOLD_END:
		outbuf = "</B>";
		break;
	case EMPHASIS:
		outbuf = "<EM>";
		break;
	case EMPHASIS_END:
		outbuf = "</EM>";
		break;
	case ITALICS:
		outbuf = "<I>";
		break;
	case ITALICS_END:
		outbuf = "</I>";
		break;
	case PLAIN:
		outbuf = "<PL>";
		break;
	case PLAIN_END:
		outbuf = "</PL>";
		break;
	case STRONG:
		outbuf = "<S>";
		break;
	case STRONG_END:
		outbuf = "</S>";
		break;
	case TT:
		outbuf = "<TT>";
		break;
	case TT_END:
		outbuf = "</TT>";
		break;
	case UNDERLINE:
		outbuf = "<U>";
		break;
	case UNDERLINE_END:
		outbuf = "</U>";
		break;
	case SIZE_0:
		outbuf = "<0>";
		break;
	case SIZE_0_END:
		outbuf = "</0>";
		break;
	case SIZE_1:
		outbuf = "<1>";
		break;
	case SIZE_1_END:
		outbuf = "</1>";
		break;
	case SIZE_2:
		outbuf = "<2>";
		break;
	case SIZE_2_END:
		outbuf = "</2>";
		break;
	case SIZE_3:
		outbuf = "<3>";
		break;
	case SIZE_3_END:
		outbuf = "</3>";
		break;
	case SIZE_4:
		outbuf = "<4>";
		break;
	case SIZE_4_END:
		outbuf = "</4>";
		break;
	case SIZE_5:
		outbuf = "<5>";
		break;
	case SIZE_5_END:
		outbuf = "</5>";
		break;
	case SIZE_6:
		outbuf = "<6>";
		break;
	case SIZE_6_END:
		outbuf = "</6>";
		break;
	case SIZE_7:
		outbuf = "<7>";
		break;
	case SIZE_7_END:
		outbuf = "</7>";
		break;
	case SIZE_8:
		outbuf = "<8>";
		break;
	case SIZE_8_END:
		outbuf = "</8>";
		break;
	case SIZE_9:
		outbuf = "<9>";
		break;
	case SIZE_9_END:
		outbuf = "</9>";
		break;

	case RED:
		outbuf = "<r>";
		break;
	case RED_END:
		outbuf = "</r>";
		break;
	case GREEN:
		outbuf = "<g>";
		break;
	case GREEN_END:
		outbuf = "</g>";
		break;
	case BLUE:
		outbuf = "<b>";
		break;
	case BLUE_END:
		outbuf = "</b>";
		break;
	case CYAN:
		outbuf = "<c>";
		break;
	case CYAN_END:
		outbuf = "</c>";
		break;
	case MAGENTA:
		outbuf = "<m>";
		break;
	case MAGENTA_END:
		outbuf = "</m>";
		break;
	case YELLOW:
		outbuf = "<y>";
		break;
	case YELLOW_END:
		outbuf = "</y>";
		break;
	case BLACK:
		outbuf = "<k>";
		break;
	case BLACK_END:
		outbuf = "</k>";
		break;
	case WHITE:
		outbuf = "<w>";
		break;
	case WHITE_END:
		outbuf = "</w>";
		break;
	default:
		myError("Bad node type in NodeToSMLNG");
	}
	if(outbuf != NULL)
		fwrite(outbuf, sizeof(char), strlen(outbuf), fp);
}



/* convert our node list back to SML/NG */
void NodeListToSMLNG(NODE *head, FILE *fp)
{
	while(head != NULL)
	{
		NodeToSMLNG(head, fp);
		head = head->next;
	}
}

void *safeMalloc(size_t size)
{
    void * ret = malloc(size);
    if(ret == NULL)
        myError("safeMalloc failed");
    return ret;
}


NODESTK *push(NODESTK *oldStk, NODE *item)
{
    NODESTK *stk = (NODESTK *)safeMalloc(sizeof(NODESTK));
    stk->item = item;
    stk->down = oldStk;
    return stk;
}

NODE *pop(NODESTK **stk)
{
    NODE *ret;
    NODESTK *newStk = (*stk)->down;
    if(*stk == NULL)
        myError("Popping from empty node stack");

    ret = (*stk)->item;
    free(*stk);
    *stk = newStk;  
    return ret;
}

NODE *newNode()
{
    NODE *node = (NODE *)safeMalloc(sizeof(NODE));
    memset(node, 0, sizeof(NODE));
    node->in.size = -1;
    return node;
}

NODE *GetTagNode(char *fileBuf, int bufLen, int *handle)
{
    NODE *node = newNode();
    int closingTag = 0;    /* is this a closing tag */

    (*handle)++;

    if(fileBuf[*handle] == '/')
    {
        closingTag = 1;
        (*handle)++;
    }

    if(fileBuf[*handle] == 'B')        
        node->type = ( closingTag? BOLD_END: BOLD);
    else if(!memcmp(&fileBuf[*handle], "EM", 2))        
        node->type = ( closingTag? EMPHASIS_END: EMPHASIS);
    else if(fileBuf[*handle] == 'I')        
        node->type = ( closingTag? ITALICS_END: ITALICS);
    else if(!memcmp(&fileBuf[*handle], "PL", 2))        
        node->type = ( closingTag? PLAIN_END: PLAIN);
    else if(fileBuf[*handle] == 'S')        
        node->type = ( closingTag? STRONG_END: STRONG);
    else if(!memcmp(&fileBuf[*handle], "TT", 2))        
        node->type = ( closingTag? TT_END: TT);
    else if(fileBuf[*handle] == 'U')        
        node->type = ( closingTag? UNDERLINE_END: UNDERLINE);
    else if(fileBuf[*handle] == '0')        
        node->type = ( closingTag? SIZE_0_END: SIZE_0);
    else if(fileBuf[*handle] == '1')        
        node->type = ( closingTag? SIZE_1_END: SIZE_1);
    else if(fileBuf[*handle] == '2')        
        node->type = ( closingTag? SIZE_2_END: SIZE_2);
    else if(fileBuf[*handle] == '3')        
        node->type = ( closingTag? SIZE_3_END: SIZE_3);
    else if(fileBuf[*handle] == '4')        
        node->type = ( closingTag? SIZE_4_END: SIZE_4);
    else if(fileBuf[*handle] == '5')        
        node->type = ( closingTag? SIZE_5_END: SIZE_5);
    else if(fileBuf[*handle] == '6')        
        node->type = ( closingTag? SIZE_6_END: SIZE_6);
    else if(fileBuf[*handle] == '7')        
        node->type = ( closingTag? SIZE_7_END: SIZE_7);
    else if(fileBuf[*handle] == '8')        
        node->type = ( closingTag? SIZE_8_END: SIZE_8);
    else if(fileBuf[*handle] == '9')        
        node->type = ( closingTag? SIZE_9_END: SIZE_9);
    else if(fileBuf[*handle] == 'r')        
        node->type = ( closingTag? RED_END: RED);
    else if(fileBuf[*handle] == 'g')        
        node->type = ( closingTag? GREEN_END: GREEN);
    else if(fileBuf[*handle] == 'b')        
        node->type = ( closingTag? BLUE_END: BLUE);
    else if(fileBuf[*handle] == 'c')        
        node->type = ( closingTag? CYAN_END: CYAN);
    else if(fileBuf[*handle] == 'm')        
        node->type = ( closingTag? MAGENTA_END: MAGENTA);
    else if(fileBuf[*handle] == 'y')        
        node->type = ( closingTag? YELLOW_END: YELLOW);
    else if(fileBuf[*handle] == 'k')        
        node->type = ( closingTag? BLACK_END: BLACK);
    else if(fileBuf[*handle] == 'w')        
        node->type = ( closingTag? WHITE_END: WHITE);
    else
        myError("Wrong Tag");

    while(*handle < bufLen && fileBuf[*handle] != '>')
        (*handle)++;

    if(fileBuf[*handle] == '>')
        (*handle)++;

    return node;
}


/* the scanner */
NODE *scan(char *fileBuff, int bufLen, int *handle)
{
    NODE *node = NULL;
    int handle1 = *handle;
    int textSize;

    /* have we come to the end */
    if(*handle == bufLen)
        return NULL;

    /* see if this is a tag */
    if(fileBuff[*handle] == '<')
    {
        node = GetTagNode(fileBuff, bufLen, handle);
        return node;
    }    

    /* no this is text */
    while(handle1 < bufLen && fileBuff[handle1] != '<')
        handle1++;

    textSize = handle1 - *handle;
    
    node = newNode();

    node->type = TEXT;

    node->s = safeMalloc(sizeof(char) * (textSize+1));
    memcpy(node->s, &fileBuff[*handle], textSize);
    node->s[textSize] = '\0';

    *handle = handle1;

    return node;
}

CONTEXT AddTagToContext(CONTEXT *in, NODE *node)
{
    CONTEXT ctxt = *in;

    switch(node->type)
    {
    case TEXT:
        break;

    case BOLD:
        ctxt.b = 1;		
		ctxt.n_b = node;
        break;

    case EMPHASIS:
        if(!ctxt.s)
            ctxt.em = !ctxt.em;
        break;

    case ITALICS:
        ctxt.i = 1;
		ctxt.n_i = node;
        break;

    case PLAIN:
        ctxt.u = ctxt.s = ctxt.em = ctxt.i = ctxt.b = ctxt.tt = 0;
		ctxt.n_u1 = ctxt.n_u2 = ctxt.n_u3 = ctxt.n_s = ctxt.n_i = ctxt.n_b = ctxt.n_tt = NULL;
        break;

    case STRONG:
        ctxt.s = 1;
		ctxt.n_s = node;
        ctxt.em = 0;
        break;

    case TT:
        ctxt.tt = 1;
		ctxt.n_tt = node;
        break;

    case UNDERLINE:
        if(ctxt.u < 3)
		{
            ctxt.u++;
			switch(ctxt.u)
			{
			case 1:
				ctxt.n_u1 = node;
				break;
			case 2:
				ctxt.n_u2 = node;
				break;
			case 3:
				ctxt.n_u3 = node;
				break;
			default:
				myError("In add tag to context, wrong value of underline");
			}
		}
        break;

    case SIZE_0:
        ctxt.size = 0;
		ctxt.n_size = node;
        break;

    case SIZE_1:
        ctxt.size = 1;
		ctxt.n_size = node;
        break;

    case SIZE_2:
        ctxt.size = 2;
		ctxt.n_size = node;
        break;

    case SIZE_3:
        ctxt.size = 3;
		ctxt.n_size = node;
        break;

    case SIZE_4:
        ctxt.size = 4;
		ctxt.n_size = node;
        break;

    case SIZE_5:
        ctxt.size = 5;
		ctxt.n_size = node;
        break;

    case SIZE_6:
        ctxt.size = 6;
		ctxt.n_size = node;
        break;

    case SIZE_7:
        ctxt.size = 7;
		ctxt.n_size = node;
        break;

    case SIZE_8:
        ctxt.size = 8;
		ctxt.n_size = node;
        break;

    case SIZE_9:
        ctxt.size = 9;
		ctxt.n_size = node;
        break;

    case RED:
        ctxt.color = RED;
		ctxt.n_color = node;
        break;

    case GREEN:
        ctxt.color = GREEN;
		ctxt.n_color = node;
        break;

    case BLUE:
        ctxt.color = BLUE;
		ctxt.n_color = node;
        break;

    case CYAN:
        ctxt.color = CYAN;
		ctxt.n_color = node;
        break;

    case MAGENTA:
        ctxt.color = MAGENTA;
		ctxt.n_color = node;
        break;
        
    case YELLOW:
        ctxt.color = YELLOW;
		ctxt.n_color = node;
        break;

    case BLACK:
        ctxt.color = BLACK;
		ctxt.n_color = node;
        break;

    case WHITE:
        ctxt.color = WHITE;
		ctxt.n_color = node;
        break;
        
    default:
        myError("AddTagToContext called with bad type");
    }
    return ctxt;
}



NODE * parse()
{
    char *fileBuff = NULL;    /* a buffer of 5 MB will be allocated here, and will be freed at the end of parsing */
    int bufLen;    /* the number of characters in this buffer */
    
    NODE *oldNode = NULL, *node = NULL, *headNode = NULL; /* headNode is the one that'll be returned */

    NODESTK *stk = NULL;    /* node stack */
    int handle = 0;    /* for the scanner, has to be zero at the start */

    /*** Initialization and allocation ***/
    /* allocate the initial fileBuff */
    fileBuff = (char *)safeMalloc(FIVE_MB * sizeof(char));

    /* read the entire program into the buffer */
    bufLen = fread(fileBuff, sizeof(char), FIVE_MB, stdin);
    if(bufLen == 0)
    {
        free(fileBuff);
        return NULL;
    }

    /*** start parsing ***/
    headNode = node = scan(fileBuff, bufLen, &handle);
    if(headNode == NULL)
    {
        /* nothing could be parsed */
        myError("No input");
        return NULL;
    }
    while(node != NULL)
    {
        /* set the prev next pointers */
        node->prev = oldNode;
        if(oldNode != NULL)        
            oldNode->next = node;
        

        /* now do the other node type specific things */
        if(isstarttag(node->type))        
        {
            /* push node on stack for use with the "match" pointer */
            stk = push(stk, node);
        }
        else if(isendtag(node->type))
        {
            NODE *matchNode = pop(&stk);
            node->match = matchNode;
            matchNode->match = node;
        }
        else if(node->type == TEXT)
        {
            /* if this is a text node, set the prev_text, and next_text, pointers */ 
            NODE *p = node->prev;
            while(p != NULL && p->type != TEXT)
                p = p->prev;

            if(p != NULL)
            {
                node->prevtext = p;
                p->nexttext = node;
            }

            /* DOUBT: what to do about prev tag and next tag. What exactly do we want there? */
        }
        else
        { 
            myError("Garbage in node->type\n");
        }
    
        if(node->prev != NULL)
        {
            /*
            if previous tag is an ending tag, the incoming context for this, is the same as the incoming context,
            for the matching opening tag for that ending tag 
            "if it is a closing tag, restore the context that was saved at the corresponding open tag"
            */
            if(isendtag(node->prev->type))
                    node->in = node->prev->match->in;
            else /* previous is an opening tag */
                node->in = AddTagToContext(&node->prev->in, node->prev);
        }
        
        /* get set for the next iteration */
        oldNode = node;
        node = scan(fileBuff, bufLen, &handle);
    }

        free(fileBuff);
    return headNode;
}

char *ShowContext(CONTEXT *ctxt)
{
    static char buf[256];
    sprintf(buf, "(b:%d, em:%d, i:%d, s:%d, tt:%d, u:%d, size:%d, color:%d)",
        ctxt->b,
        ctxt->em,
        ctxt->i,
        ctxt->s,
        ctxt->tt,
        ctxt->u,
        ctxt->size,
        ctxt->color);
    return buf;
}


void ShowNode(NODE *node, FILE *fp)
{
    fprintf(fp, 
        "Node:%x: type:%d, prev:%x, next:%x, s:%s, prevtext:%x, nexttext:%x, match:%x, ctxt:%s\n", 
        node,
        node->type,
        node->prev,
        node->next,
        node->s==NULL?"<null>":node->s,
        node->prevtext,
        node->nexttext,
        node->match,
        ShowContext(&node->in)
    );
}

ShowNodeList(NODE *head, FILE *fp)
{
    while(head != NULL)
    {
        ShowNode(head, fp);
        head = head->next;
    }
}

int main()
{
    NODE *head = parse();
	NODE *head1 = CopyNodeList(head);    

	head = Dep_GarbageCollect(head);
	WhiteSpaceOptimization(head);
	head = EM_elim(head);
/*
	head = PL_ShortCut( head );
*/
	head = rem_redundant(head);
    NodeListToSMLNG(head, stdout);
}
