#include <stdio.h>
#include "typedefs.h"

NODE *EM_elim(NODE *root)
{
    /* Eliminate EM */

    /*
     * (1) If EM appears when S is already present, remove EM and /EM
     * (2) If we have <EM><EM> ... </EM></EM>, remove both pairs
     *     provided there is no text between them - other tags are ok and
     *     we already know from (1) that there is no <S> between the
     *     two EMs.
     */

    NODE *match, *next, *node, *match1, *match2, *first, *tnode, *second;
    int text_in_betw;

    /* (1) */

    node = root;
    while (node != NULL)
    {
        next = node -> next;
        if (node -> type == EMPHASIS)
        {
            if ((node -> in).s)
            {
                /* EM when s is already 1 */

                match = node -> match;
                if (next == match)
                {
                    /* we have <EM></EM> - move 1 more to the right */

                    next = match -> next;
                }

                root = List_remove(root, match);
                root = List_remove(root, node);
            }
        }
        node = next;
    }

    /* (2) */

    node = root;
    first = NULL;
    while (node != NULL)
    {
        next = node -> next;
        if (node -> type == TEXT || node -> type == EMPHASIS_END)
        {
            first = NULL;         /* reset - rule not applicable */
        }
        else if (node -> type == EMPHASIS)
        {
            if (first == NULL)
            {
                first = node;
            }
            else
            {
                /* this is the second <EM> with no intervening text or </EM> */

                second = node;

                /*
                 * check if there is any intervening text between the
                 * corresponding </EM> for both
                 */

                match1 = first -> match;
                match2 = second -> match;

                tnode = match2;
                text_in_betw = 0;
                while (tnode != match1)
                {
                    tnode = tnode -> next;
                    if (tnode -> type == TEXT)
                    {
                        text_in_betw = 1;
                        break;
                    }
                }
                if (!text_in_betw)
                {
                    /*
                     * we can delete first, second, match1, and match2
                     * but before we do that, next should point to the
                     * next remaining node
                     */

                    if (next == match2) next = next -> next;
                    if (next == match1) next = next -> next;

                    root = List_remove(root, match1);
                    root = List_remove(root, match2);
                    root = List_remove(root, first);
                    root = List_remove(root, second);
                }
            }
        }
        node = next;
    }
    return root;
}


NODE *rem_redundant(NODE *root)
{
    /* remove redundant tags */

    /*
     * (1) If a tag sets something that is already set, remove the tag+end
     * (2) If we have <X></X> for any X remove both
     * (3) If we have </X><X> for any X, remove both
     */

    NODE *node, *next, *tnode, *match;

    /* (2) */

    node = root;
    while (node != NULL)
    {
        next = node -> next;
        if (node -> type != TEXT &&
            isstarttag(node -> type) &&
            (node -> match == node -> next))
        {
            if (node -> prev != NULL)
            {
                /* go back; we may find more such redundancies */

                next = node -> prev;
            }
            else
            {
                /* can't go back - we are at the beginning */

                next = (node -> match) -> next;
            }
            root = List_remove(root, node -> match);
            root = List_remove(root, node);
        }
        node = next;
    }

    /* (3) */

    node = root;
    while (node != NULL)
    {
        next = node -> next;
        if (next != NULL &&
            node -> type != TEXT &&
            ((node -> type == BOLD_END && next -> type == BOLD) ||
             (node -> type == EMPHASIS_END && next -> type == EMPHASIS) ||
             (node -> type == ITALICS_END && next -> type == ITALICS) ||
             (node -> type == PLAIN_END && next -> type == PLAIN) ||
             (node -> type == STRONG_END && next -> type == STRONG) ||
             (node -> type == TT_END && next -> type == TT) ||
             (node -> type == SIZE_0_END && next -> type == UNDERLINE) ||
             (node -> type == SIZE_1_END && next -> type == SIZE_0) ||
             (node -> type == SIZE_2_END && next -> type == SIZE_2) ||
             (node -> type == SIZE_3_END && next -> type == SIZE_3) ||
             (node -> type == SIZE_4_END && next -> type == SIZE_4) ||
             (node -> type == SIZE_5_END && next -> type == SIZE_5) ||
             (node -> type == SIZE_6_END && next -> type == SIZE_6) ||
             (node -> type == SIZE_7_END && next -> type == SIZE_7) ||
             (node -> type == SIZE_8_END && next -> type == SIZE_8) ||
             (node -> type == SIZE_9_END && next -> type == SIZE_9) ||
             (node -> type == RED_END && next -> type == RED) ||
             (node -> type == GREEN_END && next -> type == GREEN) ||
             (node -> type == BLUE_END && next -> type == BLUE) ||
             (node -> type == CYAN_END && next -> type == CYAN) ||
             (node -> type == MAGENTA_END && next -> type == MAGENTA) ||
             (node -> type == YELLOW_END && next -> type == YELLOW) ||
             (node -> type == BLACK_END && next -> type == BLACK) ||
             (node -> type == WHITE_END && next -> type == WHITE))
           )
        {
            /* node and next are to be deleted */

            tnode = next;
            next = tnode -> next;
            root = List_remove(root, node);
            root = List_remove(root, tnode);
        }
        node = next;
    }

    /* (1) */

    node = root;
    while (node != NULL)
    {
        next = node -> next;
        if (next != NULL &&
            isstarttag(node -> type) &&
            isequal(node -> in, next -> in))
        {
            /* this is a beginning tag and it does not change the context */

            match = node -> match;
            if (next == match)
            {
                /* we have <X></X> - move 1 more to the right */

                next = match -> next;
            }
            root = List_remove(root, match);
            root = List_remove(root, node);
        }
        node = next;
    }

    return root;
}
