#include "typedefs.h"

/* function to remove and deallocate node from the list beginning at head */
/* if this node is of type TEXT, the prev text and next text pointers will be updated */
/* the function returns the head of the changed list */
NODE *List_remove(NODE *head, NODE *node)
{
	if(node->prev)
	{
		node->prev->next = node->next;
	}
	else
		head = node->next; /* node == head */

	if(node->next)
		node->next->prev = node->prev;

	if(node->type == TEXT)
	{
		if(node->prevtext)
			node->prevtext->nexttext = node->nexttext;
		if(node->nexttext)
			node->nexttext->prevtext = node->prevtext;
	}

	free(node);
	return head;
}

/* function to remove a tag and its match from the list */
/* if this is a text node, nothing is done */
/* it returns the new list */
NODE *RemoveTagPair(NODE *head, NODE *node)
{
	NODE *match;
	
	if(node->type == TEXT)	
		return head;

	match = node->match;
	head = List_remove(head, node);
	head = List_remove(head, match);
	return head;
}

int isequal(CONTEXT c1, CONTEXT c2)
{
    /* compares two contexts */

    return (c1.b == c2.b && c1.em == c2.em && c1.i == c2.i &&
            c1.s == c2.s && c1.tt == c2.tt && c1.u == c2.u &&
            c1.size == c2.size && c1.color == c2.color);
}


/* gives the current space context */
CONTEXT SpaceContext(CONTEXT ctxt)
{	
	ctxt.s = ctxt.em = ctxt.i = ctxt.b = 0;
	if(ctxt.u == 0)
		ctxt.color = WHITE;
	return ctxt;
}